#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

class Reader
{
protected:
	FILE *f;
	int putBackCh;
public:
	Reader(FILE *F) : putBackCh(-1), f(F) { assert(f); }
	~Reader() { if (f) fclose(f); }
	void PutBack(char ch) { assert(putBackCh < 0); putBackCh = ch; }
	int GetCh() {
		if (putBackCh >= 0) { int ch = putBackCh; putBackCh = -1; return ch; }
		else { /*printf("!");*/ return fgetc(f); }}
	bool SkipWs(int& ch) { bool any = false; for ( ; ; ) {
		ch = GetCh();
//printf("<%d; %d>", any, ch);
		if (ch == EOF) return any;
		else if (! (isspace(ch) || ch == '\xa' || ch == '\xd')) { PutBack(ch); return any; }
		else any = true; } }
	bool SkipSpaces(int& ch) { bool any = false; for ( ; ; ) {
		ch = GetCh();
		if (ch == EOF) return any;
		else if (! (isspace(ch) && ch != '\xa' && ch != '\xd')) { PutBack(ch); return any; }
		else any = true; } }
	bool ReadInt(int& i, FILE *fRept, const int lineNo, const char *cesa);
	bool ReadString(char *dest, int maxChars, const char* extraStopChars); // stop at space or extraStopChars
	bool AtEof() { int ch = GetCh(); if (ch < 0) return true; else { PutBack(ch); return false; }}
	bool AtEol() { int ch = GetCh(); bool b = (ch == '\xa' || ch == '\xd');
		if (ch >= 0) PutBack(ch);
		return b; }
	bool ReadLn() {
		int ch = GetCh(); if (ch < 0) return false;
		if (ch == '\xa') return true;
		if (ch == '\xd') {
			ch = GetCh();
			if (ch >= 0 && ch != '\xa') PutBack(ch);
			return true; }
		PutBack(ch); return false; }
	void ReadLnA() { bool ok = ReadLn(); assert(ok); }
};

bool Reader::ReadString(char *dest, int maxChars, const char* extraStopChars)
{
	int nRead = 0;
	for ( ; ; )
	{
		int ch = GetCh();
		if (ch < 0 || isspace(ch) || strchr(extraStopChars, ch)) { dest[nRead] = 0; if (ch > 0) PutBack(ch); return true; }
		if (nRead == maxChars) { dest[nRead] = 0; return false; }
		dest[nRead++] = char(ch);
	}
}

bool Reader::ReadInt(int& i, FILE *fRept, const int lineNo, const char *cesa)
{
	i = 0;
	int nDigits = 0;
	bool negative = false;
	{int ch = GetCh(); if (ch == '-') negative = true; else PutBack(ch);}
	for ( ; ; )
	{
		int ch = GetCh();
		if (ch < 0) { if (nDigits <= 0) { fprintf(fRept, "Napaka pri branju %s v %d. vrstici.  Pricakoval bi stevilo, nasel sem EOF.\n", cesa, lineNo); return false; }
			else return true; }
		if (ch < int('0') || ch > int('9')) { PutBack(ch);
			if (nDigits <= 0) { fprintf(fRept, "Napaka pri branju %s v %d. vrstici.  Pricakoval bi stevilo, nasel sem znak 0x%02x.\n", cesa, lineNo, int(ch)); return false; }
			else return true; }
		// 2^31 = 2147483648
		//printf("i = %d\n", i);
		if (i > 214748364) { fprintf(fRept, "Napaka pri branju %s v %d. vrstici.  Stevilo je preveliko (overflow).\n", cesa, lineNo); return false; }
		if (i < -214748364) { fprintf(fRept, "Napaka pri branju %s v %d. vrstici.  Stevilo je premajhno (underflow).\n", cesa, lineNo); return false; }
		/*
		int j = 2147483647 - i, d = int(ch - '0');
		//printf("i = %d, j = %d, d = %d\n", i, j, d);
		if (j < d) { fprintf(fRept, "%d. stevilo je preveliko (overflow).\n", katero); return false; }
		*/
		int d = int(ch - '0');
		if (i == 214748364 && d > 7) { fprintf(fRept, "Napaka pri branju %s v %d. vrstici.  Stevilo je preveliko (overflow).\n", cesa, lineNo); return false; }
		if (i == -214748364 && d > 8) { fprintf(fRept, "Napaka pri branju %s v %d. vrstici.  Stevilo je premajhno (underflow).\n", cesa, lineNo); return false; }
		i *= 10;
		if (negative) i -= d; else i += d;
		nDigits++;
	}
}

const int MaxLabelLen = 20;
const int MaxProgLen = 100000;
const int MaxRunLen = 1000000;
const int MinRegNo = 1;
const int MaxRegNo = 256;

class TRec
{
public:
	char key[MaxLabelLen + 1];
	int address, hc2;
	TRec *next;
	TRec(const char *Key, int Address, int Hc2, TRec* Next) :
		address(Address), next(Next), hc2(Hc2) { strcpy(key, Key); }
	~TRec() { if (next) delete next; }
};

class TInstruction
{
public:
	bool bReg, cReg;
	int a, b, c, lineNo;
	TRec* d;     // jump to d if b - c < 0; can be 0
	TRec* label; // of this instruction; can be 0

	TInstruction() : bReg(true), cReg(true), a(-1), b(-1), c(-1), d(0), lineNo(-1), label(0) { }
};

class THash
{
public:

	TRec **slots; int nSlots;

	THash() {
		nSlots = 100017;
		slots = new TRec*[nSlots];
		for (int i = 0; i < nSlots; i++) slots[i] = 0; }
	~THash() {
		for (int i = 0; i < nSlots; i++) if (slots[i]) delete slots[i];
		delete[] slots; }

	void CalcHc(const char *key, int& hc1, int& hc2) {
		hc1 = 0; hc2 = 0;
		while (*key) { int i = tolower(*key++);
			hc1 = ((hc1 * 255) + i) % nSlots;
			hc2 = ((hc1 * 123) + i) % nSlots; }}

	TRec* Access(const char *key, const bool addIfNotPresent)
	{
		int hc1, hc2; CalcHc(key, hc1, hc2);
		for (TRec* r = slots[hc1]; r; r = r->next)
			if (r->hc2 == hc2) if (stricmp(r->key, key) == 0) return r;
		if (addIfNotPresent) { slots[hc1] = new TRec(key, -1, hc2, slots[hc1]); return slots[hc1]; }
		else return 0;
	}
};

class TProgram
{
private:
	TProgram& operator= (const TProgram& ) { }
	TProgram(const TProgram& ) { }
public:
	TInstruction *inst;
	int nInst, ip;
	THash labels;
	int regs[MaxRegNo + 1];
	char *regNames[MaxRegNo + 1];
	bool lastStepWasJump;

	TProgram() {
		nInst = 0; ip = 0;
		for (int i = 0; i <= MaxRegNo; i++) regNames[i] = 0;
		inst = new TInstruction[MaxProgLen]; }
	~TProgram() { 
		for (int i = 0; i <= MaxRegNo; i++) if (regNames[i]) delete[] regNames[i];
		delete inst; }

	void ClrRegs(const int value = 2004) { for (int i = 0; i <= MaxRegNo; i++) regs[i] = value; }
	void Step();

};

void TProgram::Step()
{
	assert(0 <= ip && ip < nInst);
	TInstruction& i = inst[ip];
	regs[i.a] = (i.bReg ? regs[i.b] : i.b) - (i.cReg ? regs[i.c] : i.c);
	lastStepWasJump = false;
	if (regs[i.a] < 0 && i.d) { ip = i.d->address; lastStepWasJump = true; }
	else ip++;
}

bool TestLabel(char *p)
{
	if (! *p) return false;
	if (! isalpha(*p) && *p != '_') return false;
	while (*p) {
		if (! isalnum(*p) && *p != '_')  return false;
		p++; }
	return true;
}

bool ReadProgram(TProgram& prog, FILE *fOut, FILE *fRept)
{
	assert(MaxLabelLen >= 3); // for "sbn"
	int ch; Reader r(fOut);
	int lineNo = 1;
	for ( ; ; lineNo++)
	{
		TInstruction inst; inst.lineNo = lineNo;
		// Skip initial whitespace.
		r.SkipSpaces(ch); if (r.AtEol()) { r.ReadLnA(); continue; } if (r.AtEof()) break;
		// Read the label.
		char label[MaxLabelLen + 1];
		if (! r.ReadString(label, MaxLabelLen, ":")) { fprintf(fRept, "Predolga labela v vrstici %d.  Prvih %d znakov: \"%s\".\n", lineNo, MaxLabelLen, label);  return false; }
		// The label should be followed by a colon.
		r.SkipSpaces(ch); ch = r.GetCh();
		if (ch != int(':')) r.PutBack(ch);
		else
		{
			if (! TestLabel(label)) { fprintf(fRept, "Neustrezna labela v vrstici %d: \"%s\".\n", lineNo, label);  return false; }
			TRec* rec = prog.labels.Access(label, true);
			if (rec->address >= 0) { fprintf(fRept, "Podvojena labela v vrstici %d: \"%s\".\n", lineNo, label);  return false; }
			rec->address = prog.nInst;
			inst.label = rec;
			// After the colon, there should be "SBN".
			r.SkipSpaces(ch);
			if (! r.ReadString(label, MaxLabelLen, "")) { fprintf(fRept, "Napaka v vrstici %d: pricakoval bi \"sbn\", nasel sem \"%s\".\n", lineNo, label);  return false; }
		}
		// If there was no label, 'label' now contains "SBN".
		// If there was a label, it also contains "SBN" because we just read it above.
		if (stricmp(label, "sbn") != 0) { fprintf(fRept, "Napaka v vrstici %d: pricakoval bi \"sbn\", nasel sem \"%s\".\n", lineNo, label);  return false; }
		r.SkipSpaces(ch);
		// Read the first operand.
		if (r.AtEof() || r.AtEol()) { fprintf(fRept, "Napaka v vrstici %d: manjka prvi operand.\n", lineNo);  return false; }
		ch = r.GetCh(); if (ch != 'r' && ch != 'R') { fprintf(fRept, "Napaka v vrstici %d: prvi operand bi moral biti register in se zaceti na \"r\", ne na \"%c\".\n", lineNo, char(ch));  return false; }
		if (! r.ReadInt(inst.a, fRept, lineNo, "prvega operanda")) return false;
		if (inst.a < MinRegNo || inst.a > MaxRegNo) { fprintf(fRept, "Napaka pri prvem operandu v vrstici %d: stevilka registra bi morala biti z intervala %d..%d, ne pa %d.\n", lineNo, MinRegNo, MaxRegNo, inst.a);  return false; }
		r.SkipSpaces(ch); ch = r.GetCh(); if (ch == ',') r.SkipSpaces(ch); else r.PutBack(ch);
		// Read the second operand.
		if (r.AtEof() || r.AtEol()) { fprintf(fRept, "Napaka v vrstici %d: manjka drugi operand.\n", lineNo);  return false; }
		ch = r.GetCh(); if (ch == 'r' || ch == 'R') inst.bReg = true; else { inst.bReg = false; r.PutBack(ch); }
		if (! r.ReadInt(inst.b, fRept, lineNo, "drugega operanda")) return false;
		if (inst.bReg && (inst.b < MinRegNo || inst.b > MaxRegNo)) { fprintf(fRept, "Napaka pri drugem operandu v vrstici %d: stevilka registra bi morala biti z intervala %d..%d, ne pa %d.\n", lineNo, MinRegNo, MaxRegNo, inst.b);  return false; }
		r.SkipSpaces(ch); ch = r.GetCh(); if (ch == ',') r.SkipSpaces(ch); else r.PutBack(ch);
		// Read the third operand.
		if (r.AtEof() || r.AtEol()) { fprintf(fRept, "Napaka v vrstici %d: manjka tretji operand.\n", lineNo);  return false; }
		ch = r.GetCh(); if (ch == 'r' || ch == 'R') inst.cReg = true; else { inst.cReg = false; r.PutBack(ch); }
		if (! r.ReadInt(inst.c, fRept, lineNo, "tretjega operanda")) return false;
		if (inst.cReg && (inst.c < MinRegNo || inst.c > MaxRegNo)) { fprintf(fRept, "Napaka pri tretjem operandu v vrstici %d: stevilka registra bi morala biti z intervala %d..%d, ne pa %d.\n", lineNo, MinRegNo, MaxRegNo, inst.c);  return false; }
		r.SkipSpaces(ch); ch = r.GetCh(); bool comma = (ch == ','); if (comma) r.SkipSpaces(ch); else r.PutBack(ch);
		// Read the fourth operand, if present.
		if (comma || ! (r.AtEof() || r.AtEol()))
		{
			if (! r.ReadString(label, MaxLabelLen, "")) { fprintf(fRept, "Predolga labela kot cetrti operand v vrstici %d.  Prvih %d znakov: \"%s\".\n", lineNo, MaxLabelLen, label);  return false; }
			if (! TestLabel(label)) { fprintf(fRept, "Neustrezna labela kot cetrti operand v vrstici %d: \"%s\".\n", lineNo, label);  return false; }
			inst.d = prog.labels.Access(label, true);
		}
		// There should be nothing else in this line.
		r.SkipSpaces(ch); if (! (r.AtEof() || r.AtEol())) { ch = r.GetCh(); fprintf(fRept, "Napaka v vrstici %d: za cetrtim operandom pride se nepricakovani znak \"%c\".\n", lineNo, char(ch));  return false; }
		r.ReadLn();
		if (prog.nInst >= MaxProgLen) { fprintf(fRept, "Napaka v vrstici %d: program je predolg -- to bi bila ze %d. instrukcija v programu.\n", lineNo, prog.nInst + 1);  return false; }
		//fprintf(fRept, "# SBN r%d, %s%d, %s%d", inst.a, (inst.bReg ? "r" : ""), inst.b, (inst.cReg ? "r" : ""), inst.c);
		//if (inst.d) fprintf(fRept, ", %s", inst.d->key);
		//fprintf(fRept, "\n");
		prog.inst[prog.nInst++] = inst;
	}
	// Test if all the labels that appear as the fourth operand
	// have really been defined.
	{for (int i = 0; i < prog.nInst; i++)
		if (prog.inst[i].d && prog.inst[i].d->address < 0) { fprintf(fRept, "Napaka v vrstici %d: cetrti operand je \"%s\", ki pa ni labela nobene vrstice.\n", lineNo, prog.inst[i].d->key); return false; }}
	return true;
}

// > 1 --> error
// <= 0 --> OK (# points = return value + 10)
//
// Line -> [S] [Label [S] ":"] Sbn S Reg [S] [","] [S] RegOrImm [S] [","] [S] RegOrImm [S] [[","] [S] Label] [S]
// Label -> ("_" | Alpha) ("|" | Alpha | Digit)*
// RegOrImm -> Reg | Imm
// Reg -> ("r" | "R") Digit Digit*
// Imm -> ["-"] Digit Digit*
// Alpha -> "A" | "B" | ... | "Z" | "a" | "b" | ... | "z"
// Digit -> "0" | "1" | ... | "9"
// Sbn -> ("S" | "s") ("B" | "b") ("N" | "n")
// S -> (any character c such that isspace(c) == true)*
//
// Label names are not case-sensitive.
//
// Of course, in the grammar above, we neglected the fact
// that if thw two RegOrImm's are both Imm's without any
// intervening space, commas and '-' signs, they will
// in fact be parsed as one Imm.
//

bool ParseInt(const char *p, const char *r, int& dest)
{
	if (p >= r) return false;
	bool minus = (*p == '-');
	if (*p == '-' || *p == '+') { p++; if (p >= r) return false; }
	dest = 0;
	while (p < r) {
		if (*p < '0' || *p > '9') return false;
		dest *= 10; dest += *p++ - '0'; }
	if (minus) dest = -dest;
	return true;
}

void ProcessCommand(const char *pStart, const char *pEnd, FILE *fRept, bool isSetCmd, TProgram &prog, bool *dumpFlags)
{
	const char *p = pStart; while (p < pEnd && *p != '=') *p++;
	if (isSetCmd) { if (p == pEnd) return; }
	const char *r = pStart; while (r < p && *r != '-') *r++;
	int from, to, value;
	if (r == p) {
		if (! ParseInt(pStart, p, from)) return;
		to = from; }
	else {
		if (r == p - 1) to = MaxRegNo;
		else { if (! ParseInt(r + 1, p, to)) return; }
		if (r == pStart) from = MinRegNo;
		else { if (! ParseInt(pStart, r, from)) return; } }
	char *regName = 0;
	if (isSetCmd) { if (! ParseInt(p + 1, pEnd, value)) return; }
	else {
		if (p != pEnd) {
			int len = pEnd - (p + 1); if (len > 20) len = 20;
			regName = new char[len + 1];
			for (int i = 0; i < len; i++) regName[i] = p[i + 1];
			regName[len] = '\0'; }
	}

	if (from < MinRegNo) from = MinRegNo; else if (from > MaxRegNo) from = MaxRegNo;
	if (to < MinRegNo) to = MinRegNo; else if (to > MaxRegNo) to = MaxRegNo;
	if (to < from) return;
	for (int i = from; i <= to; i++) {
		dumpFlags[i] = true;
		if (regName) {
			if (prog.regNames[i]) delete[] prog.regNames[i];
			prog.regNames[i] = new char[strlen(regName) + 1];
			strcpy(prog.regNames[i], regName); }
		if (isSetCmd) prog.regs[i] = value; }
	if (isSetCmd) {
		if (from == to) fprintf(fRept, "r%d := %d\n", from, value);
		else fprintf(fRept, "r%d, ..., r%d := %d\n", from, to, value); }
	if (regName) delete[] regName;
}

bool Test2(const char* fnOut, const char *fnRept, const char *initValues, const char *regsToDump)
{
	// Read the program.
	FILE *fRept = (fnRept ? fopen(fnRept, "wt") : stdout); assert(fRept);
	FILE *fOut = fopen(fnOut, "rb"); if (! fOut) { fprintf(fRept, "Ne morem odpreti datoteke \"%s\".\n", fnOut ? fnOut : "\0"); fclose(fRept); return false; }
	TProgram prog;
	if (! ReadProgram(prog, fOut, fRept)) { fclose(fOut); if (fnRept) fclose(fRept); return false; }
	fclose(fOut); if (fnRept) fclose(fRept);
	// Parse 'initValues'.
	prog.ClrRegs();
	bool dumpFlags[MaxRegNo + 1]; {for (int i = 0; i <= MaxRegNo; i++) dumpFlags[i] = false;}
	if (! initValues) {
		for (int i = 0; i <= MaxRegNo; i++) dumpFlags[i] = true; }
	else {
		for (const char *p = initValues; ; ) {
			const char *r = p; while (*r && *r != ';' && *r != ',') r++;
			ProcessCommand(p, r, fRept, true, prog, dumpFlags);
			if (! *r) break; else p = r + 1; } }
	// Parse 'regsToDump'.
	if (regsToDump) {
		for (int i = 0; i <= MaxRegNo; i++) dumpFlags[i] = false;
		for (const char *p = regsToDump; ; ) {
			const char *r = p; while (*r && *r != ';' && *r != ',') r++;
			ProcessCommand(p, r, fRept, false, prog, dumpFlags);
			if (! *r) break; else p = r + 1; } }
	// Run the program.
	prog.ip = 0;
	for (int stepNo = -1; stepNo < MaxRunLen; stepNo++)
	{
		if (prog.ip == prog.nInst) break;
		char buf[500];
		if (stepNo == -1) sprintf(buf, "Zacetne vrednosti:");
		else {
			enum { MaxLen = 50 }; char op1[MaxLen + 1], op2[MaxLen + 1], op3[MaxLen + 1];
			TInstruction& i = prog.inst[prog.ip];
			if (prog.regNames[i.a]) strncpy(op1, prog.regNames[i.a], MaxLen);
			else sprintf(op1, "r%d", i.a);
			if (! i.bReg) sprintf(op2, "%d", i.b);
			else if (prog.regNames[i.b]) strncpy(op2, prog.regNames[i.b], MaxLen);
			else sprintf(op2, "r%d", i.b);
			if (! i.cReg) sprintf(op3, "%d", i.c);
			else if (prog.regNames[i.c]) strncpy(op2, prog.regNames[i.c], MaxLen);
			else sprintf(op3, "r%d", i.c);
			char label[MaxLabelLen + 3]; 
			if (i.label) strncpy(label, i.label->key, MaxLabelLen + 1);
			else sprintf(label, "__line_%d", i.lineNo);
			strcat(label, ":");
			sprintf(buf, "Korak %3d] %20s  SBN %s, %s, %s", stepNo, label, op1, op2, op3); }
		if (stepNo >= 0) {
			prog.Step();
			if (prog.lastStepWasJump) { strcat(buf, " --> "); strcat(buf, prog.inst[prog.ip].label->key); strcat(buf, "   "); } }
		fprintf(fRept, "%-70s", buf);
		for (int i = MinRegNo; i <= MaxRegNo; i++) if (dumpFlags[i]) {
			if (prog.regNames[i]) fprintf(fRept, " %s=%d", prog.regNames[i], prog.regs[i]);
			else fprintf(fRept, " r%d=%d", i, prog.regs[i]); }
		fprintf(fRept, "\n");
	}
	if (prog.ip != prog.nInst) { fprintf(fRept, "Program se ne konca vedno v %d ali manj korakih.\n", MaxRunLen); if (fnRept) fclose(fRept); return false; }
	if (fnRept) fclose(fRept);
	return true;
}

int main(int argc, char **argv)
{
	if (argc < 2 || argc > 4) {
		printf("Uporaba:  SbnTest vhod.txt [zacetne vrednosti [registri za izpis]]\n");
		printf("   vhod.txt -- Datoteka z zaporedjem ukazov, ki bi ga rad simuliral.\n");
		printf("   zacetne vrednosti -- Niz ukazov oblike \"12-34=56\", locenih z\n");
		printf("        vejicami.  Tak ukaz postavi zacetne vrednosti registrov\n");
		printf("        r12, r13, ..., r33, r34 na 56.  \"12=56\" pomeni isto kot \"12-12=56\".\n");
		printf("        Ce sta ta in naslednji parameter izpuscena, so zacetne vrednosti\n");
		printf("        vseh registrov na 2004.\n");
		printf("   registri za izpis -- Niz stevilk registrov in/ali intervalov oblike \"12-34\",\n");
		printf("        ki doloca, katere registre naj program izpise po vsakem simuliranem\n");
		printf("        koraku.  Ce je ta parameter izpuscen, se izpisejo vsi, ki so jim bile\n");
		printf("        nastavljene zacetne vrednosti s prejsnjim parametrom.\n");
		printf("        Registrom lahko tudi nastavis novo ime, npr. \"12=tralala\".\n");
		printf("\n");
		printf("Primer:   SbnSimulator sbn-povprecje.txt 1=100,2=101,3=123,4=456,5=576 1-5,6=povp,7=vsota\n");
		return 0; }
	Test2(argv[1], 0, (argc > 2 ? argv[2] : 0), (argc > 3 ? argv[3] : 0));
	return 0;
}

